#include "3DShapes.h"

/*
Returns the bounce of an object of velocity v into a plain of normal n
*/
Vector rayReflection(const Vector& v, const  Vector& n) {
	return v - (2 * dot(n, v)) * n;
}

bool intersection(const Ray& ray, const Plane& plane, Scalar& k, Versor& n) {
	Scalar dn = dot(ray.dir, plane.n.asVector());
	if (dn == 0)
		return false;
	Scalar k0 = dot((plane.p - ray.start), plane.n.asVector()) / dn;
	if (k0 < 0 || k0 > k)
		return false;
	k = k0;
	n = plane.n;
	return true;
}

// Returns the first zero of ax^2 + bx + c
Scalar radix(Scalar a, Scalar b, Scalar c) {
	Scalar delta = b * b - 4 * a * c;
	if (delta < 0)
		return NO_SOLUTION;
	return (-b - sqrt(delta)) / (2 * a);
}

bool intersection(const Ray& r, const Sphere& s, Scalar& k, Versor& n) {
	Vector pc = r.start - s.center;

	Scalar A = dot(r.dir, r.dir);
	Scalar B = 2 * dot(pc, r.dir);
	Scalar C = dot(pc, pc) - (s.radius * s.radius);

	Scalar k0 = radix(A, B, C);
	if (k0 < 0 || k0 > k)
		return false;

	k = k0;
	Point hitPos = r.start + k * r.dir;
	n = normalize(hitPos - s.center);
	return true;
}