#include "3DShapes.h";
#include "image.h"

const int RES = 1000;
const Color BACKGROUND(0.1f, 0.5f, 0.9f);

/* OUR SCENE TO BE RENDERED */
/* just global vars, for simplicity */
Sphere sphere1 = { Point( 2.7,-0.8, 6.5), 2.0 };
Sphere sphere2 = { Point(-1,  -0.5, 3.5), 1.5f };
Sphere sphere3 = { Point( 0.4,-0.8, 3.5), 1.0f };
Plane plane = { Point(0, -1.4, 0), Versor::up() };
Scalar focalDist = 1;
Versor viewDir = Versor::backward();
Versor lightDir = normalize(Vector(-2, 2.5, -0.5)); // direction TOWARD the light
Image output(RES, RES);

Color applyLighting(const Versor& normal, const Color& baseColor, bool isInShadow) {
	Scalar diffusiveTerm = dot(normal, lightDir);
	Versor halfWay = mix(lightDir, viewDir, 0.5f);
	Scalar specularTerm = pow(dot(normal, halfWay), 60);

	diffusiveTerm = std::max(diffusiveTerm, 0.0f);

    if (isInShadow) {
        // no direct light (only ambient)
        diffusiveTerm = 0;
        specularTerm = 0;
    }

    const Scalar ambientTerm = 0.2f;
    return (diffusiveTerm + ambientTerm) * baseColor + specularTerm * Color(0.5f, 0.5f, 0.5f);
}

Scalar intersectAll(const Ray& ray, Versor& n, Color& c) {
	Scalar k = INFINITY;
	c = BACKGROUND;
    n = Versor::up();
    if ( intersection(ray, plane  , k, n) ) c = Color(0.1f, 0.6f, 0.0f);
    if ( intersection(ray, sphere1, k, n) ) c = Color(0.8f, 0.2f, 0.2f);
    if ( intersection(ray, sphere2, k, n) ) c = Color(1.0f, 0.5f, 0.0f);
    if ( intersection(ray, sphere3, k, n) ) c = Color(1.0f, 0.45f, 0.0f);

	return k;
}

void rayTrace() {
	for (int x = 0; x < RES; x++) {
		for (int y = 0; y < RES; y++) {
			Point pixelPos(
				x / (RES - 1.0f) * 2 - 1,
				y / (RES - 1.0f) * 2 - 1,
				focalDist
			);
			Point eyePos = Point::origin();
			Ray primary({ eyePos, pixelPos - eyePos });
			Versor normal;
			Color baseColor;
			Scalar k = intersectAll(primary, normal, baseColor);
			if (k != INFINITY) {

				Point hitPos = primary.start + (k - 0.01f) * primary.dir;
				Ray raySecondary = { hitPos, lightDir.asVector() };
                Versor unusedN;
                Color unusedC;
                Scalar k1 = intersectAll(raySecondary, unusedN, unusedC);

                bool inShadow = (k1 < INFINITY) ;

                Color c = applyLighting(normal, baseColor, inShadow);

				output.pixel(x, y) = c;
			}
			else {
				output.pixel(x, y) = BACKGROUND;
			}
		}
	}
}

int main() {
	rayTrace();

    output.exportPPM("/temp/rayTrace.ppm");
	return 0;
}
