#pragma once

#include "vector.h"
#include "quaternion.h"
#include "point.h"
#include "versor.h"

class Transform {
public:
	Scalar s; // scaling OR size
	Quaternion r; // rotation OR orientation
	Vector t; // translation OR position

	Transform() : s(1), r(1), t(0, 0, 0) {}
	Transform(Scalar _s, Quaternion _r, Vector _t) : s(_s), r(_r), t(_t) {}

	Point apply(const Point& p) const {
		return rotate(r, p.scaled(s)) + t;
	}
	Vector apply(const Vector& v) const {
		return rotate(r, v * s);
	}
	Versor apply(const Versor& v) const {
		return rotate(r, v);
	}
	Scalar apply(const Scalar len) const {
		return s * len;  // len is supposed to be a length
	}
	// Sinonimo operator() come sinonimo di apply
	// template<class T> T operator() (const T& x) const { return apply(x); }
	Point operator() (const Point& p) const { return apply(p); }
	Vector operator() (const Vector& v) const { return apply(v); }
	Versor operator() (const Versor& v) const { return apply(v); }
	Scalar operator() (const Scalar sc) const { return apply(sc); }

	Transform inverse() const {
		return Transform(
			1 / s,
			~r,
			rotate(~r, -t / s)
		);
	}

	// inversion (in-place version)
	void invert() {
		s = 1 / s;
		r.conjugate();
		t = apply(-t);
	}

	// cumulation (in-place version)
	void operator *= (const Transform& before) {
		// (first "before", then me!)
		t += apply(before.t);
		s *= before.s;
		r *= before.r;
	}
};

/*
inline void operator *= (Transform& after , const Transform& before) {
	// (first "before", then me!)
	after.t += after.apply(before.t);
	after.s *= before.s;
	after.r *= before.r;
}
*/


// cumulation of transformations
inline Transform operator*(const Transform& after, const Transform& before) {
	return Transform(
		after.s * before.s,
		after.r * before.r,
		after.t + rotate(after.r, before.t * after.s)
		// oppure
		// after.t + after.apply(before.t)
	);
}
