#pragma once
#include "vector.h"

class Versor {
public:
    Versor() : x(1), y(0), z(0) {}

    Versor(const Vector& v) {
        Scalar norm = length(v);
        this->x = v.x / norm;
        this->y = v.y / norm;
        this->z = v.z / norm;
    }

    static Versor up() { return Versor(0, +1, 0); }
    static Versor down() { return Versor(0, -1, 0); }
    static Versor forward() { return Versor(0, 0, +1); }
    static Versor backward() { return Versor(0, 0, -1); }
    static Versor right() { return Versor(+1, 0, 0); }
    static Versor left() { return Versor(-1, 0, 0); }

    Vector asVector() const { return Vector(x, y, z); }

    Vector operator* (Scalar k) const { return asVector() * k; }
    Versor operator- () const { return Versor(-x, -y, -z); }

    friend Scalar dot(const Versor& v1, const Versor& v2);

private:
    Scalar x, y, z;

    Versor(Scalar _x, Scalar _y, Scalar _z) : x(_x), y(_y), z(_z) {};
};

// Global Functions

inline Versor normalize(Vector v) {
    return Versor(v);
}

inline Scalar dot(const Versor& v1, const Versor& v2) {
    return (v1.x * v2.x) + (v1.y * v2.y) + (v1.z * v2.z);
}

inline Versor mix(const Versor& a, const Versor& b, Scalar t) {
    // N-Lerp
    return normalize(mix(a.asVector(), b.asVector(), t));
}

inline Scalar sinBetween(const Versor& a, const Versor& b) {
    return length(cross(a.asVector(), b.asVector()));
}

inline Scalar cosBetween(const Versor& a, const Versor& b) {
    return dot(a.asVector(), b.asVector());
}

inline Scalar angleBetween(const Versor& a, const Versor& b) {
    return atan2(sinBetween(a, b), cosBetween(a, b));
}
